package iottid

import (
	"crypto/tls"
	"crypto/x509"
	"encoding/base64"
	"encoding/hex"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"net/http"
	"strings"
	"sync"
	"time"

	"github.com/google/uuid"
)

const (
	// tidSvrHTTPHost IotTid服务端URL(HTTP)
	tidSvrHTTPHost = "http://tidauthrelay.iot.cloud.tencent.com:80"
	// tidSvrHTTPSHost IotTid服务端URL(HTTPS)
	tidSvrHTTPSHost = "https://tidauthrelay.iot.cloud.tencent.com:443"
	// authDeviceCgiURL 认证设备CGI地址后缀
	authDeviceCgiURL = "/cgi-bin/auth-device"
	// tidSvrCgiTimeout IotTid服务端CGI超时时间
	tidSvrCgiTimeout = time.Duration(1000 * time.Millisecond)
	// deviceDataBytesLen 设备端认证字节流长度
	deviceDataBytesLen = 132
)

// requestAuthDevice 认证设备CGI请求结构
type requestAuthDevice struct {
	RequestID       string `json:"RequestId"`
	Version         string `json:"Version"`
	Timestamp       int    `json:"Timestamp"`
	ProtocolVersion string `json:"ProtocolVersion"`
	CipherSuite     string `json:"CipherSuite"`
	Tid             string `json:"Tid"`
	DeviceRandom    string `json:"DeviceRandom"`
	DevicePubKey    string `json:"DevicePubKey"`
	DeviceSign      string `json:"DeviceSign"`
	Uin             string `json:"Uin"`
	Signature       string `json:"Signature"`
}

// responseAuthDevice 认证设备CGI返回结构
type responseAuthDevice struct {
	Retcode int    `json:"Retcode"`
	Retmsg  string `json:"Retmsg"`
	Result  struct {
		ProtocolVersion string `json:"ProtocolVersion"`
		CipherSuite     string `json:"CipherSuite"`
		ServerPubKey    string `json:"ServerPubKey"`
		ServerRandom    string `json:"ServerRandom"`
		ServerSign      string `json:"ServerSign"`
		Token           string `json:"Token"`
		Signature       string `json:"Signature"`
	} `json:"Result"`
}

// checkTidSvrSignature 校验TID服务器签名
func checkTidSvrSignature(cgiResponse *responseAuthDevice, tidSvrPublicKeyPath string) error {
	originData := fmt.Sprintf(
		"CipherSuite=%s&ProtocolVersion=%s&ServerPubKey=%s&ServerRandom=%s&ServerSign=%s&Token=%s",
		cgiResponse.Result.CipherSuite,
		cgiResponse.Result.ProtocolVersion,
		cgiResponse.Result.ServerPubKey,
		cgiResponse.Result.ServerRandom,
		cgiResponse.Result.ServerSign,
		cgiResponse.Result.Token,
	)

	signatureBytes, e := base64.StdEncoding.DecodeString(cgiResponse.Result.Signature)
	if e != nil {
		return fmt.Errorf("Decode signature error: %s", e)
	}
	tidSvrPublicKey, e := ioutil.ReadFile(tidSvrPublicKeyPath)
	if e != nil {
		return fmt.Errorf("Read server public key file error: %s", e)
	}
	e = RsaVerifySignWithSha256([]byte(originData), signatureBytes, []byte(tidSvrPublicKey))
	if e != nil {
		return fmt.Errorf("Verify signature error: %s", e.Error())
	}
	return nil
}

// generateAuthDeviceUserSignature 生成用户签名
func generateAuthDeviceUserSignature(cgiRequest *requestAuthDevice, userPrivateKey string) (string, error) {
	originData := fmt.Sprintf(
		"CipherSuite=%s&DevicePubKey=%s&DeviceRandom=%s&DeviceSign=%s&ProtocolVersion=%s&RequestId=%s&Tid=%s&Timestamp=%d&Uin=%s&Version=%s",
		cgiRequest.CipherSuite,
		cgiRequest.DevicePubKey,
		cgiRequest.DeviceRandom,
		cgiRequest.DeviceSign,
		cgiRequest.ProtocolVersion,
		cgiRequest.RequestID,
		cgiRequest.Tid,
		cgiRequest.Timestamp,
		cgiRequest.Uin,
		cgiRequest.Version,
	)

	signatureBytes, e := RsaSignWithSha256([]byte(originData), []byte(userPrivateKey))
	if e != nil {
		return "", fmt.Errorf("Generate signature error: %s", e.Error())
	}
	signature := base64.StdEncoding.EncodeToString(signatureBytes)
	return signature, nil
}

// generateAuthDeviceServerData 生成服务端认证字节流
func generateAuthDeviceServerData(cgiResponse *responseAuthDevice) ([]byte, error) {
	var serverData []byte
	var bs []byte
	var e error

	if bs, e = hex.DecodeString(cgiResponse.Result.ProtocolVersion); e != nil {
		return nil, fmt.Errorf("Tid server cgi return value [ProtocolVersion] invalid")
	}
	serverData = append(serverData, bs...)
	if bs, e = hex.DecodeString(cgiResponse.Result.CipherSuite); e != nil {
		return nil, fmt.Errorf("Tid server cgi return value [CipherSuite] invalid")
	}
	serverData = append(serverData, bs...)
	if bs, e = hex.DecodeString(cgiResponse.Result.ServerRandom); e != nil {
		return nil, fmt.Errorf("Tid server cgi return value [ServerRandom] invalid")
	}
	serverData = append(serverData, bs...)
	if bs, e = hex.DecodeString(cgiResponse.Result.ServerPubKey); e != nil {
		return nil, fmt.Errorf("Tid server cgi return value [ServerPubKey] invalid")
	}
	serverData = append(serverData, bs...)
	if bs, e = hex.DecodeString(cgiResponse.Result.ServerSign); e != nil {
		return nil, fmt.Errorf("Tid server cgi return value [ServerSign] invalid")
	}
	serverData = append(serverData, bs...)

	return serverData, nil
}

// generateAuthDeviceRequest 生成CGI请求包
func generateAuthDeviceRequest(userUin string, deviceData []byte) *requestAuthDevice {
	cgiRequest := requestAuthDevice{
		RequestID:       uuid.New().String(),
		Version:         "1.0",
		Timestamp:       int(time.Now().Unix()),
		ProtocolVersion: hex.EncodeToString(deviceData[:1]),
		CipherSuite:     hex.EncodeToString(deviceData[1:2]),
		Tid:             hex.EncodeToString(deviceData[2:18]),
		DeviceRandom:    hex.EncodeToString(deviceData[18:34]),
		DevicePubKey:    hex.EncodeToString(deviceData[34:100]),
		DeviceSign:      hex.EncodeToString(deviceData[100:132]),
		Uin:             userUin,
	}
	return &cgiRequest
}

var (
	httpClientInitOnce  sync.Once
	httpsClientInitOnce sync.Once
	httpClient          *http.Client
	httpsClient         *http.Client
)

// getHTTPClient 获取HTTP客户端对象
func getHTTPClient() *http.Client {
	httpClientInitOnce.Do(func() {
		httpClient = &http.Client{
			Timeout: tidSvrCgiTimeout,
		}
	})
	return httpClient
}

// getHTTPSClient 获取HTTPS客户端对象
func getHTTPSClient(tidSvrCert []byte) *http.Client {
	httpsClientInitOnce.Do(func() {
		caCertPool := x509.NewCertPool()
		caCertPool.AppendCertsFromPEM(tidSvrCert)
		httpsClient = &http.Client{
			Transport: &http.Transport{
				TLSClientConfig: &tls.Config{
					RootCAs: caCertPool,
				},
			},
			Timeout: tidSvrCgiTimeout,
		}
	})
	return httpsClient
}

// callAuthDevice 调用认证接口并解析返回包
func callAuthDevice(cgiRequest *requestAuthDevice, client *http.Client, url string) (*responseAuthDevice, error) {
	requestData, _ := json.Marshal(*cgiRequest)
	httpResponse, e := client.Post(url, "application/json", strings.NewReader(string(requestData)))
	if e != nil {
		return nil, fmt.Errorf("Call iottid server error: %s", e.Error())
	}
	defer httpResponse.Body.Close()
	body, e := ioutil.ReadAll(httpResponse.Body)
	if e != nil {
		return nil, fmt.Errorf("Read tid server response error: %s", e.Error())
	}

	cgiResponse := responseAuthDevice{Retcode: -1}
	e = json.Unmarshal(body, &cgiResponse)
	if e != nil {
		return nil, fmt.Errorf("Decode tid server response error: %s", e.Error())
	}
	if cgiResponse.Retcode != 0 {
		return nil, fmt.Errorf("Iottid server cgi return failed: %s", cgiResponse.Retmsg)
	}

	return &cgiResponse, nil
}

// parseAuthDeviceToken 解析会话TOKEN
func parseAuthDeviceToken(cgiResponse *responseAuthDevice, userPrivateKey string) ([]byte, error) {
	encryptedTokenBytes, e := base64.StdEncoding.DecodeString(cgiResponse.Result.Token)
	if e != nil {
		return nil, fmt.Errorf("Iottid server cgi return value [Token] invalid")
	}
	token, e := RsaDecrypt(encryptedTokenBytes, []byte(userPrivateKey))
	if e != nil {
		return nil, fmt.Errorf("Iottid server cgi return value [Token] invalid")
	}
	return token, nil
}

// AuthDevice 认证设备
// Param:userUin 用户UIN
// Param:userPrivateKey 用户私钥
// Param:deviceData 设备端的认证字节流
// Return:serverData 服务端的认证字节流
// Return:token 会话token
func AuthDevice(userUin string, userPrivateKey string, deviceData []byte, tidSvrPublicKeyPath string) (serverData []byte, token []byte, err error) {
	// 参数校验
	if len(deviceData) != deviceDataBytesLen {
		err = fmt.Errorf("Device data length error")
		return
	}
	userUin = strings.TrimSpace(userUin)
	if len(userUin) == 0 {
		err = fmt.Errorf("User uin error")
		return
	}
	if len(userPrivateKey) == 0 {
		err = fmt.Errorf("User private key error")
		return
	}

	// 构造请求包
	cgiRequest := generateAuthDeviceRequest(userUin, deviceData)
	if cgiRequest.Signature, err = generateAuthDeviceUserSignature(cgiRequest, userPrivateKey); err != nil {
		return
	}

	// 发送请求
	var cgiResponse *responseAuthDevice
	if cgiResponse, err = callAuthDevice(cgiRequest, getHTTPClient(), tidSvrHTTPHost+authDeviceCgiURL); err != nil {
		return
	}

	// 校验签名
	if err = checkTidSvrSignature(cgiResponse, tidSvrPublicKeyPath); err != nil {
		return
	}

	// 解密会话密钥
	if token, err = parseAuthDeviceToken(cgiResponse, userPrivateKey); err != nil {
		return
	}

	// 构造服务端的认证字节流
	if serverData, err = generateAuthDeviceServerData(cgiResponse); err != nil {
		return
	}

	return
}

// AuthDeviceHTTPS 认证设备（HTTPS通道）
// Param:userUin 用户UIN
// Param:userPrivateKey 用户私钥
// Param:deviceData 设备端的认证字节流
// Return:serverData 服务端的认证字节流
// Return:token 会话token
func AuthDeviceHTTPS(userUin string, userPrivateKey string, deviceData []byte, tidSvrCertificatePath string) (serverData []byte, token []byte, err error) {
	// 参数校验
	if len(deviceData) != deviceDataBytesLen {
		err = fmt.Errorf("Device data length error")
		return
	}
	userUin = strings.TrimSpace(userUin)
	if len(userUin) == 0 {
		err = fmt.Errorf("User uin error")
		return
	}
	if len(userPrivateKey) == 0 {
		err = fmt.Errorf("User private key error")
		return
	}

	// 构造请求包
	cgiRequest := generateAuthDeviceRequest(userUin, deviceData)
	if cgiRequest.Signature, err = generateAuthDeviceUserSignature(cgiRequest, userPrivateKey); err != nil {
		return
	}

	// 发送请求
	var cgiResponse *responseAuthDevice
	var tidSvrCert []byte
	tidSvrCert, err = ioutil.ReadFile(tidSvrCertificatePath)
	if err != nil {
		return
	}
	if cgiResponse, err = callAuthDevice(cgiRequest, getHTTPSClient(tidSvrCert), tidSvrHTTPSHost+authDeviceCgiURL); err != nil {
		return
	}

	// 解密会话密钥
	if token, err = parseAuthDeviceToken(cgiResponse, userPrivateKey); err != nil {
		return
	}

	// 构造服务端的认证字节流
	if serverData, err = generateAuthDeviceServerData(cgiResponse); err != nil {
		return
	}

	return
}
