package iottid

import (
	"fmt"

	"github.com/tjfoc/gmsm/sm4"
)

const (
	// SM4KeyBytesLen SM4算法密钥字节数
	SM4KeyBytesLen = 16
	// SM4DataModBytesLen SM4算法明文与密文字节模数
	SM4DataModBytesLen = 16
)

// EncryptDataWithSM4 SM4加密数据
// Param:originData 业务原始数据，必须是16字节的整数倍
// Param:sessionKey 会话密钥，必须是16字节
// Return:encryptedData 加密后的数据，长度与原始数据一致
func EncryptDataWithSM4(originData []byte, sessionKey []byte) ([]byte, error) {
	if len(sessionKey) < SM4KeyBytesLen {
		return nil, fmt.Errorf("SM4 key length error")
	}
	if len(originData)%SM4DataModBytesLen != 0 {
		return nil, fmt.Errorf("Origin data length error")
	}
	block, err := sm4.NewCipher(sessionKey[:SM4KeyBytesLen])
	if err != nil {
		return nil, fmt.Errorf("Generate SM4 cipher block error")
	}
	num := len(originData) / SM4DataModBytesLen
	encryptedData := make([]byte, len(originData))
	tmpEncryptedData := make([]byte, SM4DataModBytesLen)
	for i := 0; i < num; i++ {
		block.Encrypt(tmpEncryptedData, originData[i*SM4DataModBytesLen:(i+1)*SM4DataModBytesLen])
		copy(encryptedData[i*SM4DataModBytesLen:(i+1)*SM4DataModBytesLen], tmpEncryptedData[:])
	}
	return encryptedData, nil
}

// DecryptDataWithSM4 SM4解密数据
// Param:encryptedData 加密后的数据，必须是16字节的整数倍
// Param:sessionKey 会话密钥，必须是16字节
// Return:originData 业务原始数据，长度与加密数据一致
func DecryptDataWithSM4(encryptedData []byte, sessionKey []byte) ([]byte, error) {
	if len(sessionKey) < SM4KeyBytesLen {
		return nil, fmt.Errorf("SM4 key length error")
	}
	if len(encryptedData)%SM4DataModBytesLen != 0 {
		return nil, fmt.Errorf("Encrypted data length error")
	}
	block, err := sm4.NewCipher(sessionKey[:SM4KeyBytesLen])
	if err != nil {
		return nil, fmt.Errorf("Generate SM4 cipher block error")
	}
	num := len(encryptedData) / SM4DataModBytesLen
	originData := make([]byte, len(encryptedData))
	tmpOriginData := make([]byte, SM4DataModBytesLen)
	for i := 0; i < num; i++ {
		block.Decrypt(tmpOriginData, encryptedData[i*SM4DataModBytesLen:(i+1)*SM4DataModBytesLen])
		copy(originData[i*SM4DataModBytesLen:(i+1)*SM4DataModBytesLen], tmpOriginData[:])
	}
	return originData, nil
}
