<?php

namespace IotTid;

use Exception;

/**
 * IotTid服务端操作类
 */
class Server {

    // 设备端认证字节流长度
    const DeviceDataBytesLen = 132;
    
    // IotTid服务端URL
    const HttpServerHost = 'http://tidauthrelay.iot.cloud.tencent.com:80';
    const HttpsServerHost = 'https://tidauthrelay.iot.cloud.tencent.com:443';
    
    // IotTid服务端公钥文件
    // const ServerPublicKeyPath = __DIR__ . '/server_publickey.pem';
    
    // IotTid服务端证书文件
    // const ServerCertificatePath = __DIR__ . '/server_certificate.pem';
    
    // IotTid服务端CGI地址后缀
    const AuthDeviceCgiSuffix = '/cgi-bin/auth-device';
    
    // IotTid服务端CGI超时时间
    const ServerCgiTimeout = 1.0;

    /**
     * 通过HTTP协议认证设备，获取会话TOKEN
     * @param string $userUin               用户在腾讯云注册的UIN
     * @param string $userPrivateKey        用户的私钥，必须与在腾讯云TID控制台填写的公钥匹配
     * @param string $deviceData            设备端认证字节流
     * @param string $serverPublicKeyPath   服务端公钥文件地址
     * @return array                        包含两个元素的数组，第一个元素为服务端认证字节流，第二个元素为会话TOKEN
     */
    public function authDevice($userUin, $userPrivateKey, $deviceData, $serverPublicKeyPath) {
        // 参数检查
        if (!is_string($deviceData) || strlen($deviceData) != self::DeviceDataBytesLen) {
            throw new Exception('Device data length error');
        }
        if (!is_string($userUin) || trim($userUin) == '') {
            throw new Exception('User uin error');
        }
        if (trim($userPrivateKey) == '') {
            throw new Exception('User private key error');
        }

        // 构造请求参数
        $cgiRequest = $this->generateAuthDeviceCgiRequest($userUin, $deviceData);

        // 用户服务端签名
        $cgiRequest['Signature'] = $this->generateAuthDeviceUserSignature($cgiRequest, $userPrivateKey);

        // 发送请求至IotTid服务端
        $httpClient = $this->getHttpClient();
        $authDeviceCgiURL = self::HttpServerHost . self::AuthDeviceCgiSuffix;
        $httpResult = $httpClient->post($authDeviceCgiURL, [
            'headers' => [
                'Content-Type' => 'application/json',
                'Accept' => 'application/json',
            ],
            'timeout' => self::ServerCgiTimeout,
            'body' => json_encode($cgiRequest),
        ]);
        if ($httpResult->getStatusCode() != 200) {
            throw new Exception('Call IotTid server error');
        }
        $httpResultBody = $httpResult->getBody();

        // 解析IotTid服务端返回包
        $cgiResponse = json_decode($httpResultBody, true);
        if ($cgiResponse == NULL) {
            throw new Exception(('Decode IotTid server response error'));
        }
        if (!isset($cgiResponse['Retcode']) || !is_int($cgiResponse['Retcode']) ||
            $cgiResponse['Retcode'] != 0) {
            throw new Exception('IotTid server cgi return failed');
        }

        // 验证IotTid服务端签名
        $signBuf = sprintf(
            'CipherSuite=%s&ProtocolVersion=%s&ServerPubKey=%s&ServerRandom=%s&ServerSign=%s&Token=%s',
            strval($cgiResponse['Result']['CipherSuite']),
            strval($cgiResponse['Result']['ProtocolVersion']),
            strval($cgiResponse['Result']['ServerPubKey']),
            strval($cgiResponse['Result']['ServerRandom']),
            strval($cgiResponse['Result']['ServerSign']),
            strval($cgiResponse['Result']['Token'])
        );
        if (!$this->verifyServerSignature($signBuf, strval($cgiResponse['Result']['Signature']), $serverPublicKeyPath)) {
            throw new Exception('IotTid server signature invalid');
        }

        // 解析会话TOKEN
        $token = $this->parseAuthDeviceSessionToken($cgiResponse, $userPrivateKey);

        // 构造服务端的认证字节流
        $serverData = $this->generateAuthDeviceServerData($cgiResponse);

        return [$serverData, $token];
    }

    /**
     * 通过HTTPS协议认证设备，获取会话TOKEN
     * @param string $userUin               用户在腾讯云注册的UIN
     * @param string $userPrivateKey        用户的私钥，必须与在腾讯云TID控制台填写的公钥匹配
     * @param string $deviceData            设备端认证字节流
     * @param string $serverCertificatePath 服务端证书文件地址
     * @return array                        包含两个元素的数组，第一个元素为服务端认证字节流，第二个元素为会话TOKEN
     */
    public function authDeviceHttps($userUin, $userPrivateKey, $deviceData, $serverCertificatePath) {
        // 参数检查
        if (!is_string($deviceData) || strlen($deviceData) != self::DeviceDataBytesLen) {
            throw new Exception('Device data length error');
        }
        if (!is_string($userUin) || trim($userUin) == '') {
            throw new Exception('User uin error');
        }
        if (trim($userPrivateKey) == '') {
            throw new Exception('User private key error');
        }

        // 构造请求参数
        $cgiRequest = $this->generateAuthDeviceCgiRequest($userUin, $deviceData);

        // 用户服务端签名
        $cgiRequest['Signature'] = $this->generateAuthDeviceUserSignature($cgiRequest, $userPrivateKey);

        // 发送请求至IotTid服务端
        $httpsClient = $this->getHttpsClient();
        $authDeviceCgiURL = self::HttpsServerHost . self::AuthDeviceCgiSuffix;
        $httpResult = $httpsClient->post($authDeviceCgiURL, [
            'headers' => [
                'Content-Type' => 'application/json',
                'Accept' => 'application/json',
            ],
            'timeout' => self::ServerCgiTimeout,
            'body' => json_encode($cgiRequest),
            'verify' => $serverCertificatePath,
        ]);
        if ($httpResult->getStatusCode() != 200) {
            throw new Exception('Call IotTid server error');
        }
        $httpResultBody = $httpResult->getBody();

        // 解析IotTid服务端返回包
        $cgiResponse = json_decode($httpResultBody, true);
        if ($cgiResponse == NULL) {
            throw new Exception(('Decode IotTid server response error'));
        }
        if (!isset($cgiResponse['Retcode']) || !is_int($cgiResponse['Retcode']) ||
            $cgiResponse['Retcode'] != 0) {
            throw new Exception('IotTid server cgi return failed');
        }

        // 解析会话TOKEN
        $token = $this->parseAuthDeviceSessionToken($cgiResponse, $userPrivateKey);

        // 构造服务端的认证字节流
        $serverData = $this->generateAuthDeviceServerData($cgiResponse);

        return [$serverData, $token];
    }

    /**
     * 生成认证设备CGI请求包
     * @param string $userUin           用户在腾讯云注册的UIN
     * @param string $userPrivateKey    用户的私钥，必须与在腾讯云TID控制台填写的公钥匹配
     * @param string $deviceData        设备端认证字节流
     * @return array                    CGI请求包（不含签名）
     */
    private function generateAuthDeviceCgiRequest($userUin, $deviceData) {
        $cgiRequest = [
            'RequestId' => self::uuidV4(),
            'Version' => '1.0',
            'Timestamp' => intval(time()),
            'ProtocolVersion' => bin2hex(substr($deviceData, 0, 1)),
            'CipherSuite' => bin2hex(substr($deviceData, 1, 1)),
            'Tid' => bin2hex(substr($deviceData, 2, 16)),
            'DeviceRandom' => bin2hex(substr($deviceData, 18, 16)),
            'DevicePubKey' => bin2hex(substr($deviceData, 34, 66)),
            'DeviceSign' => bin2hex(substr($deviceData, 100, 32)),
            'Uin' => $userUin,
        ];
        return $cgiRequest;
    }

    /**
     * 获取HTTP客户端实例
     * @return mixed    客户端实例
     */
    private function getHttpClient() {
        static $httpClient = false;
        if ($httpClient === false) {
            $httpClient = new \GuzzleHttp\Client();
        }
        return $httpClient;
    }

    /**
     * 获取HTTPS客户端实例
     * @return mixed    客户端实例
     */
    private function getHttpsClient() {
        static $httpsClient = false;
        if ($httpsClient === false) {
            $httpsClient = new \GuzzleHttp\Client();
        }
        return $httpsClient;
    }

    /**
     * 解析用户私钥
     * @param string $userPrivateKey    用户私钥字符串
     * @return mixed                    私钥对象
     */
    private function parseUserPrivateKey($userPrivateKey) {
        static $privKey = false;
        if ($privKey == false) {
            $privKey = \openssl_pkey_get_private($userPrivateKey);
            if (!$privKey) {
                throw new Exception('Parse user private key error');
            }
        }
        return $privKey;
    }

    /**
     * 解析IotTid服务端公钥
     * @param string $serverPublicKeyPath   服务端公钥文件地址
     * @return mixed                        公钥对象
     */
    public function parseServerPublicKey($serverPublicKeyPath) {
        static $pubKey = false;
        if ($pubKey == false) {
            $pubKey = \openssl_pkey_get_public(file_get_contents($serverPublicKeyPath));
            if (!$pubKey) {
                throw new Exception('Parse server public key error');
            }
        }
        return $pubKey;
    }

    /**
     * 针对认证设备CGI的请求包生成用户签名
     * @param string $cgiRequest        请求包
     * @param string $userPrivateKey    用户私钥字符串
     * @return string                   使用BASE64编码的签名串
     */
    private function generateAuthDeviceUserSignature($cgiRequest, $userPrivateKey) {
        $signBuf = sprintf(
            'CipherSuite=%s&DevicePubKey=%s&DeviceRandom=%s&DeviceSign=%s&ProtocolVersion=%s&RequestId=%s&Tid=%s&Timestamp=%d&Uin=%s&Version=%s',
            $cgiRequest['CipherSuite'],
            $cgiRequest['DevicePubKey'],
            $cgiRequest['DeviceRandom'],
            $cgiRequest['DeviceSign'],
            $cgiRequest['ProtocolVersion'],
            $cgiRequest['RequestId'],
            $cgiRequest['Tid'],
            $cgiRequest['Timestamp'],
            $cgiRequest['Uin'],
            $cgiRequest['Version']
        );

        $privKey = $this->parseUserPrivateKey($userPrivateKey);
        $signature = '';
        if (!openssl_sign($signBuf, $signature, $privKey, OPENSSL_ALGO_SHA256)) {
            throw new Exception('Generate user signature error');
        }
        return \base64_encode($signature);
    }

    /**
     * 用BASE64解码并校验IotTid服务端签名
     * @param string $originData        原始字符串
     * @param string $signature         使用BASE64编码的签名串
     * @return bool                     是否验签成功
     */
    private function verifyServerSignature($originData, $signature, $serverPublicKeyPath) {
        $signature = base64_decode($signature, true);
        if (!$signature) {
            return false;
        }
        $pubKey = $this->parseServerPublicKey($serverPublicKeyPath);
        $ok = openssl_verify($originData, $signature, $pubKey, OPENSSL_ALGO_SHA256);
        if ($ok == -1) {
            throw new Exception('Verify server signature error');
        }
        return $ok == 1 ? true : false;
    }

    /**
     * 解密IotTid服务端回传的数据
     * @param string $encryptedData     已加密的字符串
     * @param string $userPrivateKey    用户私钥字符串
     * @return string                   解密后的字符串
     */
    private function decryptServerData($encryptedData, $userPrivateKey) {
        $privKey = $this->parseUserPrivateKey($userPrivateKey);
        $decryptedData = '';
        if (!openssl_private_decrypt($encryptedData, $decryptedData, $privKey)) {
            throw new Exception('Decrypt server data error');
        }
        return $decryptedData;
    }

    /**
     * 从认证设备返回包中解析会话TOKEN
     * @param array $cgiResponse        返回包
     * @return string                   会话TOKEN
     */
    private function parseAuthDeviceSessionToken($cgiResponse, $userPrivateKey) {
        $encryptedToken = \base64_decode($cgiResponse['Result']['Token']);
        if (!$encryptedToken) {
            throw new Exception('IotTid server cgi return value [Token] invalid');
        }
        $token = $this->decryptServerData($encryptedToken, $userPrivateKey);
        return $token;
    }

    /**
     * 根据认证设备返回包生成服务端认证字节流
     * @param array $cgiResponse        返回包
     * @return array                    服务端认证字节流
     */
    private function generateAuthDeviceServerData($cgiResponse) {
        $serverData = '';
        $s = hex2bin($cgiResponse['Result']['ProtocolVersion']);
        if (!$s) {
            throw new Exception('IotTid server cgi return value [ProtocolVersion] invalid');
        }
        $serverData .= $s;
        $s = hex2bin($cgiResponse['Result']['CipherSuite']);
        if (!$s) {
            throw new Exception('IotTid server cgi return value [CipherSuite] invalid');
        }
        $serverData .= $s;
        $s = hex2bin($cgiResponse['Result']['ServerRandom']);
        if (!$s) {
            throw new Exception('IotTid server cgi return value [ServerRandom] invalid');
        }
        $serverData .= $s;
        $s = hex2bin($cgiResponse['Result']['ServerPubKey']);
        if (!$s) {
            throw new Exception('IotTid server cgi return value [ServerPubKey] invalid');
        }
        $serverData .= $s;
        $s = hex2bin($cgiResponse['Result']['ServerSign']);
        if (!$s) {
            throw new Exception('IotTid server cgi return value [ServerSign] invalid');
        }
        $serverData .= $s;

        return $serverData;
    }

    /**
     * 生成UUID
     * @return string       时间相关的UUID
     */
    private static function uuidV4() {
        return sprintf('%04x%04x-%04x-%04x-%04x-%04x%04x%04x',
            // 32 bits for "time_low"
            mt_rand(0, 0xffff), mt_rand(0, 0xffff),

            // 16 bits for "time_mid"
            mt_rand(0, 0xffff),

            // 16 bits for "time_hi_and_version",
            // four most significant bits holds version number 4
            mt_rand(0, 0x0fff) | 0x4000,

            // 16 bits, 8 bits for "clk_seq_hi_res",
            // 8 bits for "clk_seq_low",
            // two most significant bits holds zero and one for variant DCE1.1
            mt_rand(0, 0x3fff) | 0x8000,

            // 48 bits for "node"
            mt_rand(0, 0xffff), mt_rand(0, 0xffff), mt_rand(0, 0xffff)
        );
      }

}