# -*- coding: utf8 -*-

import uuid
import time
import json
import base64
import requests
from iottid import rsahelper

# TID服务端域名
TID_SVR_HOST = 'http://tidauthrelay.iot.cloud.tencent.com'
TID_SVR_HTTPSHOST = "https://tidauthrelay.iot.cloud.tencent.com:443"

# TID服务认证设备CGI地址
AUTH_DEVICE_CGI_URL = TID_SVR_HOST + '/cgi-bin/auth-device'
AUTH_DEVICE_CGI_URL_HTTPS = TID_SVR_HTTPSHOST + '/cgi-bin/auth-device'

# TID服务认证设备CGI超时时间
AUTH_DEVICE_CGI_TIMEOUT = 1.0

# 设备认证字节流长度
DEVICE_DATA_BYTES_LEN = 132

# http 长连接
session = None


def AuthDevice(userUin, userPrivateKey, deviceData, tidSvrPublicKeyFile):
    """认证设备，获取会话密钥及服务端认证字节流
    :param userUin: 用户UIN，类型为str
    :param userPrivateKey: 用户私钥，类型为str
    :param deviceData: 设备端认证字节流，类型为bytes
    :param tidSvrPublicKeyFile: 服务端公钥文件，类型为str
    :return: 服务端认证字节流及会话密钥，类型为(bytes, bytes)
    """
    if tidSvrPublicKeyFile is None or type(tidSvrPublicKeyFile) != str:
        raise TypeError('Server public key error')

    try:
        f = open(tidSvrPublicKeyFile)
        tidSvrPublicKey = f.read()
        f.close()
    except Exception as e:
        raise Exception('Read server public key file(%s) error: %s' %
                        (tidSvrPublicKeyFile % str(e)))
    return authDeviceCore(userUin=userUin,
                          userPrivateKey=userPrivateKey,
                          deviceData=deviceData,
                          tidSvrPublicKey=tidSvrPublicKey)


def AuthDeviceHttps(userUin, userPrivateKey, deviceData, cert):
    """认证设备，获取会话密钥及服务端认证字节流
    :param userUin: 用户UIN，类型为str
    :param userPrivateKey: 用户私钥，类型为str
    :param deviceData: 设备端认证字节流，类型为bytes
    :param cert: 服务证书路径，类型为str
    :return: 服务端认证字节流及会话密钥，类型为(bytes, bytes)
    """
    return authDeviceCore(userUin=userUin,
                          userPrivateKey=userPrivateKey,
                          deviceData=deviceData,
                          cert=cert)


def authDeviceCore(userUin,
                   userPrivateKey,
                   deviceData,
                   tidSvrPublicKey=None,
                   cert=None):
    """认证设备，获取会话密钥及服务端认证字节流
    :param userUin: 用户UIN，类型为str
    :param userPrivateKey: 用户私钥，类型为str
    :param deviceData: 设备端认证字节流，类型为bytes
    :param tidSvrPublicKey:(optional) 服务端公钥，类型为str
    :param cert:(optional) 服务证书路径，类型为str
    :return: 服务端认证字节流及会话密钥，类型为(bytes, bytes)
    """

    # 参数校验
    if type(deviceData) != bytes or len(deviceData) != DEVICE_DATA_BYTES_LEN:
        raise TypeError('Device data error')
    if type(userUin) != str or str.strip(userUin) == '':
        raise TypeError('User uin error')
    if type(userPrivateKey) != str or str.strip(userPrivateKey) == '':
        raise TypeError('User private key error')
    if tidSvrPublicKey is not None and (type(tidSvrPublicKey) != str
                                        or str.strip(tidSvrPublicKey) == ''):
        raise TypeError('Server Public key error')
    if cert is not None and type(cert) != str:
        raise TypeError('Cert file error')
    if cert is None and tidSvrPublicKey is None:
        raise TypeError('Need Cert file OR Server Public key')

    # 构造请求参数
    reqDict = {
        'RequestId': str(uuid.uuid1()),
        'Version': '1.0',
        'Timestamp': int(time.time()),
        'ProtocolVersion': deviceData[:1].hex(),
        'CipherSuite': deviceData[1:2].hex(),
        'Tid': deviceData[2:18].hex(),
        'DeviceRandom': deviceData[18:34].hex(),
        'DevicePubKey': deviceData[34:100].hex(),
        'DeviceSign': deviceData[100:132].hex(),
        'Uin': userUin,
    }

    # 签名
    signStr = 'CipherSuite=%s&DevicePubKey=%s&DeviceRandom=%s&DeviceSign=%s&ProtocolVersion=%s&RequestId=%s&Tid=%s&Timestamp=%d&Uin=%s&Version=%s' % ( \
        reqDict['CipherSuite'],
        reqDict['DevicePubKey'],
        reqDict['DeviceRandom'],
        reqDict['DeviceSign'],
        reqDict['ProtocolVersion'],
        reqDict['RequestId'],
        reqDict['Tid'],
        reqDict['Timestamp'],
        reqDict['Uin'],
        reqDict['Version'],
    )
    signedBytes = None
    try:
        signedBytes = rsahelper.RsaSignWithSha256(signStr.encode(),
                                                  userPrivateKey)
    except Exception as e:
        raise Exception('Rsa sign failed: %s' % str(e))
    reqDict['Signature'] = base64.b64encode(signedBytes).decode()

    # 发送请求
    header = {"Content-Type": "application/json"}
    try:
        if cert is None:
            response = session.post(url=AUTH_DEVICE_CGI_URL,
                                    data=json.dumps(reqDict),
                                    headers=header)
        else:
            response = session.post(url=AUTH_DEVICE_CGI_URL_HTTPS,
                                    data=json.dumps(reqDict),
                                    headers=header,
                                    verify=cert)
        if response.status_code != 200:
            raise Exception('status code %s' % str(response.status_code))
    except Exception as e:
        raise Exception('Send http request to tid server failed: %s' % str(e))

    # 解析返回包
    # respStr = respBytes.decode()
    respDict = response.json()
    if 'Retcode' not in respDict.keys() or respDict['Retcode'] != 0:
        raise Exception('Tid server handle failed')
    if 'Result' not in respDict.keys() or type(respDict['Result']) != dict:
        raise Exception('Tid server result invalid')
    for key in ('ProtocolVersion', 'CipherSuite', 'ServerPubKey',
                'ServerRandom', 'ServerSign', 'Token', 'Signature'):
        if key not in respDict['Result'].keys() or type(
                respDict['Result'][key]) != str:
            raise Exception('Tid server result invalid')

    # 校验签名
    if tidSvrPublicKey is not None:
        signStr = 'CipherSuite=%s&ProtocolVersion=%s&ServerPubKey=%s&ServerRandom=%s&ServerSign=%s&Token=%s' % ( \
            respDict['Result']['CipherSuite'],
            respDict['Result']['ProtocolVersion'],
            respDict['Result']['ServerPubKey'],
            respDict['Result']['ServerRandom'],
            respDict['Result']['ServerSign'],
            respDict['Result']['Token'],
        )
        verifyResult = rsahelper.RsaVerifySignWithSha256(
            base64.b64decode(respDict['Result']['Signature']),
            signStr.encode(), tidSvrPublicKey)
        if not verifyResult:
            raise Exception('Tid server signautre check failed')

    # 解密会话密钥
    sessionTokenBytes = None
    try:
        sessionTokenBytes = rsahelper.RsaDecrypt(
            base64.b64decode(respDict['Result']['Token']), userPrivateKey)
    except Exception as e:
        raise Exception('Rsa decrypt session token failed: %s' % str(e))

    # 构造服务端的认证字节流
    serverBytes = bytes.fromhex(respDict['Result']['ProtocolVersion']) + \
        bytes.fromhex(respDict['Result']['CipherSuite']) + \
        bytes.fromhex(respDict['Result']['ServerRandom']) + \
        bytes.fromhex(respDict['Result']['ServerPubKey']) + \
        bytes.fromhex(respDict['Result']['ServerSign'])

    return (serverBytes, sessionTokenBytes)


if __name__ != '__main__':
    session = requests.Session()
